from lxml import etree
import os
from glob import glob
from copy import deepcopy
from .xml import remove_whitespace, prettify
from IPython.display import display, Markdown
from pathlib import Path
from lxml import etree

# namespaces
tei_namespace = "http://www.tei-c.org/ns/1.0"
xml_namespace = "http://www.w3.org/XML/1998/namespace"
ns = {  
    "tei": tei_namespace,
    "xml": xml_namespace 
}
id_attrib = f'{{{xml_namespace}}}id'

def add_id_to_bibl(dir_path):
    for file_path in glob(f'{dir_path}/*.xml'):
        print(f' - Processing {file_path}')
        n = 1
        tree = etree.parse(file_path)
        for element in tree.xpath('*//tei:bibl', namespaces=ns):
            if 'n' in element.attrib:
                del element.attrib['n']
            element.attrib[id_attrib] = f'bibl-{str(n)}'
            n += 1
        pretty_tree = etree.tostring(tree, pretty_print=True, encoding='utf-8')
        with open(file_path, 'wb') as file:
            file.write(pretty_tree)

def remove_encoding_declaration(xml_string):
    return xml_string.replace('<?xml version="1.0" encoding="UTF-8"?>', '')

def find_biblstruct(bibl_struct_tree, id):
     # find correspondent biblStruct data
    xpath_expr= f"//tei:biblStruct[substring(@source, string-length(@source) - {len(id) - 1})='{id}']"
    bibl_struct_matches = bibl_struct_tree.xpath(xpath_expr, namespaces=ns)
    if len(bibl_struct_matches) > 0:
        # found it - make a copy and remove unneeded attributes
        bibl_struct_copy = deepcopy(bibl_struct_matches[0])
        if bibl_struct_copy.attrib['source']:
            del bibl_struct_copy.attrib['source']
        # add it to target listBibl
        return bibl_struct_copy
    else:
        raise RuntimeError(f"Could not find matching biblStruct element with id '{id}'.")


def create_gold_standard(bibl_content, biblstruct_content, verbose=False):
    # get source trees and elements
    bibl_tree = etree.fromstring(remove_encoding_declaration(bibl_content))
    bibl_struct_tree = etree.fromstring(remove_encoding_declaration(biblstruct_content))

    # create output document
    root = etree.Element("dataset")

    # xpath expressions to find the parent node under which the <bibl> live
    bibl_parent_xpath_list = [ 
        '//tei:note[not(tei:p)] | //tei:note[tei:p]/tei:p',
        'listBibl' 
    ]
    
    process_log = []
    
    for bibl_parent_xpath in bibl_parent_xpath_list:

        # iterate over elements, either listBibl, note, or p
        for parent_element in bibl_tree.xpath(bibl_parent_xpath, namespaces=ns):

            parent_tag = etree.QName(parent_element).localname
            parent_type = parent_element.attrib['type']
            
            # new dataset instance
            instance = etree.SubElement(root, 'instance')
            
            # input 
            child_xml = etree.tostring(parent_element, method="text", encoding='utf-8').decode()
            xml_content_text = remove_whitespace(child_xml)

            footnote_number = None
            if parent_type == "footnote":
                footnote_number = parent_element.attrib['n']
            if footnote_number:
                xml_content_text = f'{footnote_number} {xml_content_text}'
            
            input_raw = etree.SubElement(instance, 'input', {'type': 'raw'})
            input_raw.text = etree.CDATA(xml_content_text)
            input_formatted = etree.SubElement(instance, 'input', {'type': 'formatted'})
            input_formatted.text = xml_content_text

            # output
            output_bibl = etree.SubElement(instance, 'output', {'type': 'bibl'})
            output_bibl_p = etree.SubElement(output_bibl,'p', {"xmlns": tei_namespace})
            output_biblstruct = etree.SubElement(instance, 'output', {'type': 'biblStruct'})
            output_listBibl = etree.SubElement(output_biblstruct, 'listBibl', {"xmlns": tei_namespace})

            # iterate over all the contained elements, either bibl or seg
            for elem in parent_element:
                 
                elem_tag = etree.QName(elem).localname
                elem_type = elem.attrib['type'] if 'type' in elem.attrib else None

                if elem_tag == "bibl":
                    # the id of the bibl element in the document is required to look up the biblstruct
                    if id_attrib not in elem.attrib:
                        raise RuntimeError(f"Missing 'xml:id' attribute for bibl element: {elem}")
                    bibl_id = elem.attrib[id_attrib]
                    biblstruct = find_biblstruct(bibl_struct_tree, bibl_id      )           
                    # add the biblStruct
                    output_listBibl.append(biblstruct)
                
                # add the element
                output_bibl_p.append(deepcopy(elem))


    #     # node parent
    #     source_parent = bibl_element.getparent()
    #     # if the parent changes, create a new target parent
    #     if source_parent != current_source_parent:
    #         current_source_parent = source_parent
            

    #         if parent_type == "footnote" or parent_type == "inText":
    #             source_node = source_parent
    #         else:
    #             source_node = bibl_element
    #             # invalidate the pointer to the current source parent to create a new listBibl the next time
    #             current_source_parent = None

    #         attrs = {
    #             "type": parent_type,
    #         }

    #         footnote_number = None
    #         if parent_type == "footnote":
    #             footnote_number = attrs['n'] = source_node.attrib['n']

    #         # create mew listBibl element for output element
    #         current_target_parent = etree.SubElement(standoff, "listBibl", attrs)

    #         # create <desc> child to store the full input string
    #         input_full = etree.SubElement(current_target_parent, "desc", {'type': 'input-full'})

    #         # get content for it
    #         child_xml = etree.tostring(source_node, method="text", encoding='utf-8').decode()
    #         xml_content_text = remove_whitespace(child_xml)
    #         if footnote_number:
    #             xml_content_text = f'{footnote_number} {xml_content_text}'
    #         input_full.text = xml_content_text

    #         # create <desc> child to store the segmented input string
    #         input_segmented = etree.SubElement(current_target_parent, "desc", {'type': 'input-segmented'})

    #         # if bibliography, the segmented input string is the same as the raw
    #         if parent_type == "bibliography":
    #             etree.SubElement(input_segmented, "bibl").text = xml_content_text

    #         else:
    #             # Generate the high-level segments of the input string
    #             # For footnotes and in-text references, these are usually one or more <bibl> elements;
    #             # for bibliographies, there is one <bibl> element per <listBibl> element.
    #             # As there aren't any suitable other legal children elements of <desc>, we use <desc> and <bibl> to
    #             # segment the raw reference strings:
    #             # - we create a child <desc> element as direct child of the target <desc> element for each
    #             #   non-bibliographic element within the source <bibl> so that we don't have a nested structure
    #             # - we create a <bibl> element with the raw string content of the source <bibl> without the <seg type="comment"> parts
    #             for elem in source_node:

    #                 last_element = None
    #                 last_element_parts = None

    #                 for segment in elem:
    #                     segment_text_content = etree.tostring(segment, method="text", encoding='utf-8').decode().strip()

    #                     if 'type' in segment.attrib:
    #                         if segment.attrib['type'] == 'comment' or segment.attrib['type'] == 'signal':
    #                             comment_child = etree.SubElement(input_segmented, "desc")
    #                             comment_child.text = segment_text_content
    #                             last_element = None
    #                     else:
    #                         if last_element is None:
    #                             last_element = etree.SubElement(input_segmented, "bibl")
    #                             last_element_parts = [segment_text_content]
    #                         else:
    #                             last_element_parts.append(segment_text_content)
    #                         last_element.text = remove_whitespace(" ".join(last_element_parts))

    #             if footnote_number:
    #                 input_segmented[0].text = f'{footnote_number} {input_segmented[0].text}'

    #         if verbose:
    #             process_log.append(f' - Analyzing "{xml_content_text[:200]} [...]"')

    #     # find correspondent biblStruct data
    #     bibl_struct_matches = bibl_struct_tree.xpath(f"//tei:biblStruct[@n='{bibl_n}']", namespaces=ns)
    #     if len(bibl_struct_matches) > 0:
    #         # found it - make a copy and remove unneeded attributes
    #         bibl_struct_copy = deepcopy(bibl_struct_matches[0])
    #         if bibl_struct_copy.attrib['source']:
    #             del bibl_struct_copy.attrib['source']
    #         # add it to target listBibl
    #         current_target_parent.append(bibl_struct_copy)
    #     else:
    #         display(Markdown("\n".join(process_log)))
    #         raise RuntimeError(f"Could not find matching biblStruct element with id '{bibl_n}'.")

    # display(Markdown("\n".join(process_log)))
    return remove_encoding_declaration(prettify(etree.tostring(root)))


def create_all_gold_standards(bibl_dir, biblstruct_dir, biblstruct_gold_dir, verbose=False):
    for file_path in glob(f'{bibl_dir}/*.xml'):
        file_id = os.path.basename(file_path).replace(".xml", "")

        bibl_path = file_path
        biblstruct_path = f'{biblstruct_dir}/{file_id}.biblstruct.xml'
        biblstruct_gs_path = f'{biblstruct_gold_dir}/{file_id}.xml'

        # log
        md_lines = [f'### Processing {file_id}']
        md_lines.append(f'Files: [TEI/bibl]({Path(bibl_path).as_posix()}) | [TEI/biblStruct]({Path(biblstruct_path).as_posix()}) | [Gold Standard]({Path(biblstruct_gs_path).as_posix()})')
        display(Markdown("\n".join(md_lines)))

        with (open(bibl_path, 'r', encoding='utf-8') as bibl_file,
              open(biblstruct_path, 'r', encoding='utf-8') as biblStruct_file):
            bibl_content = bibl_file.read()
            biblStruct_content = biblStruct_file.read()

        output_data = create_gold_standard(bibl_content, biblStruct_content, verbose=verbose)

        with open(biblstruct_gs_path, 'w', encoding='utf-8') as output_file:
            output_file.write(output_data)