import regex as re
from lxml import etree
from IPython.display import display, Markdown
import xml.dom.minidom


def prettify(xml_string, indentation="  "):
    """Return a pretty-printed XML string"""
    return xml.dom.minidom.parseString(xml_string).toprettyxml(indent=indentation)


def remove_whitespace(text):
    # we need to remove the whitespace that comes with the indentation of pretty-printed xml
    text = re.sub(r'\n *', ' ', text)

    # reduce double spaces to one
    while re.search(r'  ', text):
        text = re.sub(r'  ', ' ', text)

    # escape character sequences which would be corrupted by whitespace removal rules
    text = re.sub(r'\. \. \.', '[!spaced_elipsis!]', text)

    # fix issues with whitespace before and after punctuation
    text = re.sub(r' ([.;,!?%:])( |$)', r'\1 ', text)
    # opening and closing punctutation, such as brackets
    text = re.sub(r'(\p{Ps}) ', r'\1', text)
    text = re.sub(r' (\p{Pe})', r'\1', text)
    # opening and closing quotes
    text = re.sub(r'(\p{Pi}) ', r'\1', text)
    text = re.sub(r' (\p{Pf})', r'\1', text)
    # slash
    text = re.sub(r' ?/ ?', r'/', text)

    # restore sequences
    text = text.replace('[!spaced_elipsis!]', r'. . .')
    return text.strip()

def fix_serialization_issues(text):
    # escape character sequences
    text = re.sub(r'\.\.\.', '[!elipsis!]', text)
    text = re.sub(r'\. \. \.', '[!spaced_elipsis!]', text)

    # remove duplicated punctuation
    text = re.sub(r'(\. ?){2}', r'.', text)

    # restore character sequences
    text = text.replace( '[!elipsis!]', '...')
    text = text.replace('[!spaced_elipsis!]', r'. . .')
    return text.strip()

def tei_to_input(tei_xml_doc):
    """
    Extract the original footnote strings from the <note> elements in a given TEI document and return a list of strings
    """
    root = etree.fromstring(tei_xml_doc)
    ref_list = []
    ns = {"tei": "http://www.tei-c.org/ns/1.0"}
    # iterate over the <note type="footnote"> and <listBibl> elements
    for xpath in ['*//tei:listBibl[@type="inText"]',
                  '*//tei:note[@type="footnote"]',
                  '*//tei:listBibl[@type="bibliography"]']:
        for element in root.findall(xpath, ns):
            localname = etree.QName(element).localname
            type = element.attrib['type'] if 'type' in element.attrib else None
            ref_parts = []
            if 'n' in element.attrib:
                ref_parts.append(element.attrib['n'])
            # iterate over the <bibl> elements
            for bibl in element.findall('tei:bibl', ns):
                # extract the text without xml tags, still contains all (collapsed) whitespace
                text = etree.tostring(bibl, method="text", encoding='utf-8').decode()
                text = remove_whitespace(text)
                ref_parts.append(text)
            if localname == "listBibl" and type != "inText":
                ref_list += ref_parts
            elif localname == "note" or (localname == "listBibl" and type == "inText"):
                ref_list.append(remove_whitespace(" ".join(ref_parts)))
    return ref_list
